/*
 * Copyright Elasticsearch B.V. and/or licensed to Elasticsearch B.V. under one
 * or more contributor license agreements. Licensed under the Elastic License
 * 2.0 and the Server Side Public License, v 1; you may not use this file except
 * in compliance with, at your election, the Elastic License 2.0 or the Server
 * Side Public License, v 1.
 */

import { EOL, hostname } from 'node:os';
import url, { URL } from 'node:url';
import type { Duration } from 'moment';
import { ByteSizeValue, offeringBasedSchema, schema, TypeOf } from '@kbn/config-schema';
import { IHttpConfig, SslConfig, sslSchema, TLS_V1_2, TLS_V1_3 } from '@kbn/server-http-tools';
import type { ServiceConfigDescriptor } from '@kbn/core-base-server-internal';
import { uuidRegexp } from '@kbn/core-base-server-internal';
import type { HttpProtocol, ICspConfig, IExternalUrlConfig } from '@kbn/core-http-server';
import type { IHttpEluMonitorConfig } from '@kbn/core-http-server/src/elu_monitor';
import type { HandlerResolutionStrategy } from '@kbn/core-http-router-server-internal';
import { CspConfig, CspConfigType } from './csp';
import { ExternalUrlConfig } from './external_url';
import {
  parseRawSecurityResponseHeadersConfig,
  securityResponseHeadersSchema,
} from './security_response_headers_config';
import { CdnConfig } from './cdn_config';
import { PermissionsPolicyConfigType } from './permissions_policy';

const SECOND = 1000;

const validBasePathRegex = /^\/.*[^\/]$/;

const hostURISchema = schema.uri({ scheme: ['http', 'https'] });
const match = (regex: RegExp, errorMsg: string) => (str: string) =>
  regex.test(str) ? undefined : errorMsg;

// The lower-case set of response headers which are forbidden within `customResponseHeaders`.
const RESPONSE_HEADER_DENY_LIST = ['location', 'refresh'];

const validHostName = () => {
  // see https://github.com/elastic/kibana/issues/139730
  return hostname().replace(/[^\x00-\x7F]/g, '');
};

/**
 * We assume the URL does not contain anything after the pathname so that
 * we can safely append values to the pathname at runtime.
 */
export const validateCdnURL = (urlString: string) => {
  const cdnURL = new URL(urlString);
  const errors: string[] = [];
  if (cdnURL.hash.length) {
    errors.push(`URL fragment not allowed, but found "${cdnURL.hash}"`);
  }
  if (cdnURL.search.length) {
    errors.push(`URL query string not allowed, but found "${cdnURL.search}"`);
  }
  if (errors.length) {
    return `CDN URL "${cdnURL.href}" is invalid:${EOL}${errors.join(EOL)}`;
  }
};

const configSchema = schema.object(
  {
    name: schema.string({ defaultValue: () => validHostName() }),
    autoListen: schema.boolean({ defaultValue: true }),
    publicBaseUrl: schema.maybe(schema.uri({ scheme: ['http', 'https'] })),
    basePath: schema.maybe(
      schema.string({
        validate: match(validBasePathRegex, "must start with a slash, don't end with one"),
      })
    ),
    shutdownTimeout: schema.duration({
      defaultValue: '30s',
      validate: (duration) => {
        const durationMs = duration.asMilliseconds();
        if (durationMs < 1000 || durationMs > 2 * 60 * 1000) {
          return 'the value should be between 1 second and 2 minutes';
        }
      },
    }),
    cdn: schema.object({
      url: schema.nullable(
        schema.maybe(schema.uri({ scheme: ['http', 'https'], validate: validateCdnURL }))
      ),
    }),
    oas: schema.object({
      enabled: schema.boolean({ defaultValue: false }),
    }),
    cors: schema.object(
      {
        enabled: schema.boolean({ defaultValue: false }),
        allowCredentials: schema.boolean({ defaultValue: false }),
        allowOrigin: schema.oneOf(
          [
            schema.arrayOf(hostURISchema, { minSize: 1 }),
            schema.arrayOf(schema.literal('*'), { minSize: 1, maxSize: 1 }),
          ],
          {
            defaultValue: ['*'],
          }
        ),
      },
      {
        validate(value) {
          if (value.allowCredentials === true && value.allowOrigin.includes('*')) {
            return 'Cannot specify wildcard origin "*" with "credentials: true". Please provide a list of allowed origins.';
          }
        },
      }
    ),
    securityResponseHeaders: securityResponseHeadersSchema,
    customResponseHeaders: schema.recordOf(schema.string(), schema.any(), {
      defaultValue: {},
      validate(value) {
        const forbiddenKeys = Object.keys(value).filter((headerName) =>
          RESPONSE_HEADER_DENY_LIST.includes(headerName.toLowerCase())
        );
        if (forbiddenKeys.length > 0) {
          return `The following custom response headers are not allowed to be set: ${forbiddenKeys.join(
            ', '
          )}`;
        }
      },
    }),
    protocol: schema.oneOf([schema.literal('http1'), schema.literal('http2')], {
      defaultValue: 'http1',
    }),
    host: schema.string({
      defaultValue: 'localhost',
      hostname: true,
    }),
    maxPayload: schema.byteSize({
      defaultValue: '1048576b',
    }),
    port: schema.number({
      defaultValue: 5601,
    }),
    rewriteBasePath: schema.boolean({ defaultValue: false }),
    ssl: sslSchema,
    keepaliveTimeout: schema.number({
      defaultValue: 120 * SECOND,
    }),
    socketTimeout: schema.number({
      defaultValue: 120 * SECOND,
    }),
    payloadTimeout: schema.number({
      defaultValue: 20 * SECOND,
    }),
    http2: schema.object({
      allowUnsecure: schema.boolean({ defaultValue: false }),
    }),
    compression: schema.object({
      enabled: schema.boolean({ defaultValue: true }),
      brotli: schema.object({
        enabled: schema.boolean({ defaultValue: false }),
        quality: schema.number({ defaultValue: 3, min: 0, max: 11 }),
      }),
      referrerWhitelist: schema.maybe(
        schema.arrayOf(
          schema.string({
            hostname: true,
          }),
          { minSize: 1 }
        )
      ),
    }),
    uuid: schema.maybe(
      schema.string({
        validate: match(uuidRegexp, 'must be a valid uuid'),
      })
    ),
    xsrf: schema.object({
      disableProtection: schema.boolean({ defaultValue: false }),
      allowlist: schema.arrayOf(
        schema.string({ validate: match(/^\//, 'must start with a slash') }),
        { defaultValue: [] }
      ),
    }),
    eluMonitor: schema.object({
      enabled: schema.boolean({ defaultValue: true }),
      logging: schema.object({
        enabled: schema.conditional(
          schema.contextRef('dist'),
          false,
          schema.boolean({ defaultValue: true }),
          schema.boolean({ defaultValue: false })
        ),
        threshold: schema.object({
          elu: schema.number({ defaultValue: 0.15, min: 0, max: 1 }),
          ela: schema.number({ defaultValue: 250, min: 0 }),
        }),
      }),
    }),
    requestId: schema.object(
      {
        allowFromAnyIp: schema.boolean({ defaultValue: false }),
        ipAllowlist: schema.arrayOf(schema.ip(), { defaultValue: [] }),
      },
      {
        validate(value) {
          if (value.allowFromAnyIp === true && value.ipAllowlist?.length > 0) {
            return `allowFromAnyIp must be set to 'false' if any values are specified in ipAllowlist`;
          }
        },
      }
    ),
    // allow access to internal routes by default to prevent breaking changes in current offerings
    restrictInternalApis: offeringBasedSchema({
      serverless: schema.boolean({ defaultValue: false }),
    }),

    versioned: schema.object({
      /**
       * Which handler resolution algo to use for public routes: "newest" or "oldest".
       *
       * @note Internal routes always require a version to be specified.
       * @note in development we have an additional option "none" which is also the default in dev.
       *       This prevents any fallbacks and requires that a version specified.
       *       Useful for ensuring that a given client always specifies a version.
       */
      versionResolution: schema.conditional(
        schema.contextRef('dev'),
        true,
        schema.oneOf([schema.literal('newest'), schema.literal('oldest'), schema.literal('none')], {
          defaultValue: 'none',
        }),
        schema.oneOf([schema.literal('newest'), schema.literal('oldest')], {
          defaultValue: 'oldest',
        })
      ),

      /**
       * Whether we require the Kibana browser build version to match the Kibana server build version.
       *
       * This number is determined when a distributable version of Kibana is built and ensures that only
       * same-build browsers can access the Kibana server.
       */
      strictClientVersionCheck: schema.boolean({ defaultValue: true }),

      /** This should not be configurable in serverless */
      useVersionResolutionStrategyForInternalPaths: offeringBasedSchema({
        traditional: schema.arrayOf(schema.string(), { defaultValue: [] }),
        serverless: schema.never(),
      }),
    }),
  },
  {
    validate: (rawConfig) => {
      if (!rawConfig.basePath && rawConfig.rewriteBasePath) {
        return 'cannot use [rewriteBasePath] when [basePath] is not specified';
      }

      if (rawConfig.publicBaseUrl) {
        const parsedUrl = url.parse(rawConfig.publicBaseUrl);
        if (parsedUrl.query || parsedUrl.hash || parsedUrl.auth) {
          return `[publicBaseUrl] may only contain a protocol, host, port, and pathname`;
        }
        if (parsedUrl.path !== (rawConfig.basePath ?? '/')) {
          return `[publicBaseUrl] must contain the [basePath]: ${parsedUrl.path} !== ${rawConfig.basePath}`;
        }
      }

      if (!rawConfig.compression.enabled && rawConfig.compression.referrerWhitelist) {
        return 'cannot use [compression.referrerWhitelist] when [compression.enabled] is set to false';
      }

      if (rawConfig.protocol === 'http2' && !rawConfig.http2.allowUnsecure) {
        const err = ensureValidTLSConfigForH2C(rawConfig.ssl);
        if (err) {
          return err;
        }
      }

      if (
        rawConfig.ssl.enabled &&
        rawConfig.ssl.redirectHttpFromPort !== undefined &&
        rawConfig.ssl.redirectHttpFromPort === rawConfig.port
      ) {
        return (
          'Kibana does not accept http traffic to [port] when ssl is ' +
          'enabled (only https is allowed), so [ssl.redirectHttpFromPort] ' +
          `cannot be configured to the same value. Both are [${rawConfig.port}].`
        );
      }
    },
  }
);

export type HttpConfigType = TypeOf<typeof configSchema>;

export const config: ServiceConfigDescriptor<HttpConfigType> = {
  path: 'server' as const,
  schema: configSchema,
  deprecations: ({ rename }) => [rename('maxPayloadBytes', 'maxPayload', { level: 'warning' })],
};

export class HttpConfig implements IHttpConfig {
  public name: string;
  public autoListen: boolean;
  public protocol: HttpProtocol;
  public host: string;
  public keepaliveTimeout: number;
  public socketTimeout: number;
  public payloadTimeout: number;
  public port: number;
  public cors: {
    enabled: boolean;
    allowCredentials: boolean;
    allowOrigin: string[];
  };
  public oas: {
    enabled: boolean;
  };
  public securityResponseHeaders: Record<string, string | string[]>;
  public customResponseHeaders: Record<string, string | string[]>;
  public maxPayload: ByteSizeValue;
  public basePath?: string;
  public publicBaseUrl?: string;
  public rewriteBasePath: boolean;
  public cdn: CdnConfig;
  public ssl: SslConfig;
  public compression: {
    enabled: boolean;
    referrerWhitelist?: string[];
    brotli: { enabled: boolean; quality: number };
  };
  public csp: ICspConfig;
  public externalUrl: IExternalUrlConfig;
  public xsrf: { disableProtection: boolean; allowlist: string[] };
  public requestId: { allowFromAnyIp: boolean; ipAllowlist: string[] };
  public versioned: {
    versionResolution: HandlerResolutionStrategy;
    strictClientVersionCheck: boolean;
    useVersionResolutionStrategyForInternalPaths: string[];
  };
  public shutdownTimeout: Duration;
  public restrictInternalApis: boolean;

  public eluMonitor: IHttpEluMonitorConfig;

  /**
   * @internal
   */
  constructor(
    rawHttpConfig: HttpConfigType,
    rawCspConfig: CspConfigType,
    rawExternalUrlConfig: ExternalUrlConfig,
    rawPermissionsPolicyConfig: PermissionsPolicyConfigType
  ) {
    this.autoListen = rawHttpConfig.autoListen;
    this.host = rawHttpConfig.host;
    this.port = rawHttpConfig.port;
    this.cors = rawHttpConfig.cors;
    const { securityResponseHeaders, disableEmbedding } = parseRawSecurityResponseHeadersConfig(
      rawHttpConfig.securityResponseHeaders,
      rawPermissionsPolicyConfig
    );
    this.securityResponseHeaders = securityResponseHeaders;
    this.customResponseHeaders = Object.entries(rawHttpConfig.customResponseHeaders ?? {}).reduce(
      (headers, [key, value]) => {
        headers[key] = Array.isArray(value)
          ? value.map((e) => convertHeader(e))
          : convertHeader(value);
        return headers;
      },
      {} as Record<string, string | string[]>
    );
    this.maxPayload = rawHttpConfig.maxPayload;
    this.name = rawHttpConfig.name;
    this.protocol = rawHttpConfig.protocol;
    this.basePath = rawHttpConfig.basePath;
    this.publicBaseUrl = rawHttpConfig.publicBaseUrl;
    this.keepaliveTimeout = rawHttpConfig.keepaliveTimeout;
    this.socketTimeout = rawHttpConfig.socketTimeout;
    this.payloadTimeout = rawHttpConfig.payloadTimeout;
    this.rewriteBasePath = rawHttpConfig.rewriteBasePath;
    this.ssl = new SslConfig(rawHttpConfig.ssl || {});
    this.compression = rawHttpConfig.compression;
    this.cdn = CdnConfig.from(rawHttpConfig.cdn);
    this.csp = new CspConfig({ ...rawCspConfig, disableEmbedding }, this.cdn.getCspConfig());
    this.externalUrl = rawExternalUrlConfig;
    this.xsrf = rawHttpConfig.xsrf;
    this.requestId = rawHttpConfig.requestId;
    this.shutdownTimeout = rawHttpConfig.shutdownTimeout;

    // default to `false` to prevent breaking changes in current offerings
    this.restrictInternalApis = rawHttpConfig.restrictInternalApis ?? false;
    this.eluMonitor = rawHttpConfig.eluMonitor;
    this.versioned = rawHttpConfig.versioned;
    this.oas = rawHttpConfig.oas;
  }
}

const convertHeader = (entry: any): string => {
  return typeof entry === 'object' ? JSON.stringify(entry) : String(entry);
};

const ensureValidTLSConfigForH2C = (tlsConfig: TypeOf<typeof sslSchema>): string | undefined => {
  if (!tlsConfig.enabled) {
    return `http2 requires TLS to be enabled. Use 'http2.allowUnsecure: true' to allow running http2 without a valid h2c setup`;
  }
  if (
    !tlsConfig.supportedProtocols.includes(TLS_V1_2) &&
    !tlsConfig.supportedProtocols.includes(TLS_V1_3)
  ) {
    return `http2 requires 'ssl.supportedProtocols' to include ${TLS_V1_2} or ${TLS_V1_3}. Use 'http2.allowUnsecure: true' to allow running http2 without a valid h2c setup`;
  }
  return undefined;
};
